-module(rumRotatorIns).

-behaviour(rumRotatorExm).

-include_lib("kernel/include/file.hrl").

-export([
   createLogFile/2
   , openLogFile/2
   , ensureLogFile/5
   , rotateLogFile/2
]).

-ifdef(TEST).
-include_lib("eunit/include/eunit.hrl").
-endif.

createLogFile(FileName, Buffer) ->
   openLogFile(FileName, Buffer).

openLogFile(FileName, Buffer) ->
   case filelib:ensure_dir(FileName) of
      ok ->
         Options =
            case Buffer of
               {Size, Interval} ->
                  [append, raw, {delayed_write, Size, Interval}];
               _ ->
                  [append, raw]
            end,
         case file:open(FileName, Options) of
            {ok, Fd} ->
               case file:read_file_info(FileName, [raw]) of
                  {ok, FileInfo} ->
                     #file_info{size = FileSize, inode = Inode} = FileInfo,
                     CTime = tryUpdateCTime(FileName, FileInfo),
                     {ok, Fd, Inode, CTime, FileSize};
                  RfiErr -> RfiErr
               end;
            OpErr -> OpErr
         end;
      EnDirErr -> EnDirErr
   end.

ensureLogFile(FileName, Fd, Inode, CTime, Buffer) ->
   case Fd of
      undefined ->
         openLogFile(FileName, Buffer);
      _ ->
         case rumUtil:isFileChanged(FileName, Inode, CTime) of
            {true, _FInfo} ->
               reopenLogfile(FileName, Fd, Buffer);
            {_, FInfo} ->
               {ok, Fd, Inode, CTime, FInfo#file_info.size}
         end
   end.

reopenLogfile(FileName, Fd, Buffer) ->
   %% Flush and close any file handles. delayed write can cause file:close not to do a close
   _ = file:datasync(Fd),
   _ = file:close(Fd),
   _ = file:close(Fd),
   openLogFile(FileName, Buffer).

%% renames failing are OK
%% IMY-TODO 文件名格式修改调整
rotateLogFile(File, 0) ->
   %% open the file in write-only mode to truncate/create it
   case file:open(File, [write]) of
      {ok, FD} ->
         _ = file:close(FD),
         _ = file:close(FD),
         tryUpdateCTime(File),
         ok;
      Error ->
         Error
   end;
rotateLogFile(File0, 1) ->
   File1 = File0 ++ ".0",
   _ = file:rename(File0, File1),
   rotateLogFile(File0, 0);
rotateLogFile(File0, Count) ->
   File1 = File0 ++ "." ++ integer_to_list(Count - 2),
   File2 = File0 ++ "." ++ integer_to_list(Count - 1),
   _ = file:rename(File1, File2),
   rotateLogFile(File0, Count - 1).

tryUpdateCTime(Name) ->
   case file:read_file_info(Name, [raw]) of
      {ok, FInfo} ->
         tryUpdateCTime(Name, FInfo);
      _ ->
         erlang:localtime()
   end.

tryUpdateCTime(Name, FileInfo) ->
   case os:type() of
      {win32, _} ->
         %注意：我们将创建时间强制为当前时间。在win32上，这可能会阻止ctime的更新：https://stackoverflow.com/q/8804342/1466825
         NewCtime = erlang:localtime(),
         NewFileInfo = FileInfo#file_info{ctime = NewCtime},
         ok = file:write_file_info(Name, NewFileInfo, [raw]),
         NewCtime;
      _ ->
         element(#file_info.ctime, FileInfo)
   end.

-ifdef(TEST).

rotate_file_test() ->
   RotCount = 10,
   {ok, TestDir} = rumUtil:create_test_dir(),
   TestLog = filename:join(TestDir, "rotation.log"),
   Outer = fun(N) ->
      ?assertEqual(ok, rumUtil:safe_write_file(TestLog, erlang:integer_to_list(N))),
      Inner = fun(M) ->
         File = lists:flatten([TestLog, $., erlang:integer_to_list(M)]),
         ?assert(filelib:is_regular(File)),
         %% check the expected value is in the file
         Number = erlang:list_to_binary(integer_to_list(N - M - 1)),
         ?assertEqual({ok, Number}, file:read_file(File))
              end,
      Count = erlang:min(N, RotCount),
      % The first time through, Count == 0, so the sequence is empty,
      % effectively skipping the inner loop so a rotation can occur that
      % creates the file that Inner looks for.
      % Don't shoot the messenger, it was worse before this refactoring.
      lists:foreach(Inner, lists:seq(0, Count - 1)),
      rotateLogFile(TestLog, RotCount)
           end,
   lists:foreach(Outer, lists:seq(0, (RotCount * 2))),
   rumUtil:delete_test_dir(TestDir).

rotate_file_zero_count_test() ->
   %% Test that a rotation count of 0 simply truncates the file
   {ok, TestDir} = rumUtil:create_test_dir(),
   TestLog = filename:join(TestDir, "rotation.log"),
   ?assertMatch(ok, rotateLogFile(TestLog, 0)),
   ?assertNot(filelib:is_regular(TestLog ++ ".0")),
   ?assertEqual(true, filelib:is_regular(TestLog)),
   ?assertEqual(1, length(filelib:wildcard(TestLog ++ "*"))),
   %% assert the new file is 0 size:
   case file:read_file_info(TestLog, [raw]) of
      {ok, FInfo} ->
         ?assertEqual(0, FInfo#file_info.size);
      _ ->
         ?assert(false)
   end,
   rumUtil:delete_test_dir(TestDir).

rotate_file_fail_test() ->
   {ok, TestDir} = rumUtil:create_test_dir(),
   TestLog = filename:join(TestDir, "rotation.log"),

   %% set known permissions on it
   ok = rumUtil:set_dir_permissions("u+rwx", TestDir),

   %% write a file
   ?assertEqual(ok, rumUtil:safe_write_file(TestLog, "hello")),

   case os:type() of
      {win32, _} -> ok;
      _ ->
         %% hose up the permissions
         ok = rumUtil:set_dir_permissions("u-w", TestDir),
         ?assertMatch({error, _}, rotateLogFile(TestLog, 10))
   end,

   %% check we still only have one file, rotation.log
   ?assertEqual([TestLog], filelib:wildcard(TestLog ++ "*")),
   ?assert(filelib:is_regular(TestLog)),

   %% fix the permissions
   ok = rumUtil:set_dir_permissions("u+w", TestDir),

   ?assertMatch(ok, rotateLogFile(TestLog, 10)),
   ?assert(filelib:is_regular(TestLog ++ ".0")),
   ?assertEqual(true, filelib:is_regular(TestLog)),
   ?assertEqual(2, length(filelib:wildcard(TestLog ++ "*"))),

   %% assert the new file is 0 size:
   case file:read_file_info(TestLog, [raw]) of
      {ok, FInfo} ->
         ?assertEqual(0, FInfo#file_info.size);
      _ ->
         ?assert(false)
   end,

   %% check that the .0 file now has the contents "hello"
   ?assertEqual({ok, <<"hello">>}, file:read_file(TestLog ++ ".0")),
   rumUtil:delete_test_dir(TestDir).

-endif.
